from ewokscore import Task
from .h5_to_spec import convert_h5


class ID22H5ToSpec(
    Task,
    input_names=["filename"],
    optional_input_names=[
        "entries",
        "outdirs",
        "primary_outdir",
        "outprefix",
        "retry_timeout",
        "rebin_filename",
        "ascii_extension",
    ],
    output_names=["outfile"],
):
    def run(self):
        entries = self.get_input_value("entries", None)
        outdirs = self.get_input_value("outdirs", None)
        primary_outdir = self.get_input_value("primary_outdir", None)
        retry_timeout = self.get_input_value("retry_timeout", 10)
        rebin_filename = self.get_input_value("rebin_filename", None)
        ascii_extension = self.get_input_value("ascii_extension", ".dat")

        self.outputs.outfile = convert_h5(
            self.inputs.filename,
            self.inputs.outprefix,
            entries=entries,
            outdirs=outdirs,
            primary_outdir=primary_outdir,
            retry_timeout=retry_timeout,
            rebin_filename=rebin_filename,
            ascii_extension=ascii_extension,
        )



import pytest
import socket
from .. import dirutils


def test_prepare_outdirs():
    primary_outdir = None
    outdirs = dict()

    result = dirutils.prepare_outdirs(outdirs, primary_outdir)
    assert result == dict()

    primary_outdir = "/users/opid22/inhouse/id222207/id22/20220701/processed"
    result = dirutils.prepare_outdirs(outdirs, primary_outdir)
    expected = {
        "primary": (
            None,
            "/users/opid22/inhouse/id222207/id22/20220701/processed",
        )
    }
    assert result == expected

    outdirs = {
        "primary": "opid22@diffract22new:/users/opid22/data1/",
        "secondary": "opid22@diffract22new:/users/opid22/data1/",
    }
    with pytest.raises(ValueError):
        dirutils.prepare_outdirs(outdirs, primary_outdir)

    outdirs = {
        "primary": f"opid22@{socket.gethostname()}:/users/opid22/data1/",
        "secondary": "opid22@diffract22new:/users/opid22/data1/",
    }
    result = dirutils.prepare_outdirs(outdirs, primary_outdir)
    expected = {
        "primary": (None, "/users/opid22/data1/"),
        "secondary": ("opid22@diffract22new", "/users/opid22/data1/"),
        "processed": (
            None,
            "/users/opid22/inhouse/id222207/id22/20220701/processed",
        ),
    }
    assert result == expected



import numpy


def unscramble_roi_collection(data, keys):
    """Sort the dictionary of numpy arrays on the keys"""
    if not keys:
        return data
    lst = [data[key] for key in keys]
    idx = list(range(len(lst[0])))
    *_, idx = zip(*sorted(zip(*lst, idx)))
    idx = list(idx)
    return {k: _apply_index(v, idx) for k, v in data.items()}


def _apply_index(v, idx):
    v = numpy.asarray(v)
    if v.ndim == 0 or v.shape[-1] != len(idx):
        return v
    return v[..., idx]



import os
import logging
from ewokscore import Task
import pyopencl
from multianalyzer.app.rebin import rebin_file
from silx.io.h5py_utils import retry, File, safe_top_level_names

from . import dirutils
from .roitools import unscramble_roi_collection

HDF5_PLUGIN_PATH = os.environ.get("HDF5_PLUGIN_PATH")
if HDF5_PLUGIN_PATH:
    del os.environ["HDF5_PLUGIN_PATH"]


logger = logging.getLogger(__name__)


@retry(retry_period=0.5, retry_timeout=10)
def read_data(filename, unscramble_method=None, entries=None, exclude_entries=None):
    res = {}
    with File(filename, mode="r") as fh:
        if entries:
            entries = set(entries)
        else:
            entries = {k for k, v in fh.items() if v.attrs.get("NX_class") == "NXentry"}
        if exclude_entries:
            entries -= set(exclude_entries)
        try:
            entries = sorted(entries, key=lambda x: float(x))
        except TypeError:
            entries = sorted(entries)
        entries = [entry for entry in entries if entry.endswith(".1")]
        print(f"Reading and unscrambling {len(entries)} entries: {entries}")
        for entry in entries:
            title = entry + "/title"
            if title not in fh:
                print(f"skip scan {entry}: no 'title'")
                continue
            end_time = entry + "/end_time"
            if end_time not in fh:
                print(f"skip scan {entry}: no 'end_time'")
                continue
            title = fh[title][()]
            try:
                title = title.decode()
            except Exception:
                pass
            if not title.startswith("fscan") and not title.startswith("f2scan"):
                print(f"skip scan {entry}: {title}")
                continue
            sort_data = {}
            non_sort_data = {}
            try:
                entry_grp = fh[entry]
                sort_data["x"] = entry_grp[
                    "instrument/eiger_roi_collection/selection/x"
                ][()]
                sort_data["y"] = entry_grp[
                    "instrument/eiger_roi_collection/selection/y"
                ][()]
                sort_data["roicol"] = entry_grp["measurement/eiger_roi_collection"][()]
                non_sort_data["arm"] = entry_grp["measurement/tth"][()]
                non_sort_data["mon"] = entry_grp["measurement/mon"][()]
                non_sort_data["tha"] = entry_grp["instrument/positioners/manom"][()]
                non_sort_data["thd"] = entry_grp["instrument/positioners/mantth"][()]
            except KeyError as e:
                logger.warning(str(e))
                continue

            if unscramble_method == "xy":
                keys = "x", "y"
            elif unscramble_method == "yx":
                keys = "y", "x"
            else:
                keys = None

            if keys:
                sort_data = unscramble_roi_collection(sort_data, keys)

                # Check the sorting
                # with File("/users/opid22/test.h5", "w") as fh:
                #    fh["x"] = sort_data["x"]
                #    fh["y"] = sort_data["y"]

            res[entry] = {**sort_data, **non_sort_data}
            for name, val in res[entry].items():
                try:
                    print(f"{name} shape = {val.shape}")
                except TypeError:
                    pass
    return res


class ID22Rebin(
    Task,
    input_names=["filename", "parsfile"],
    optional_input_names=[
        "entries",
        "debug",
        "wavelength",
        "energy",
        "step",
        "range",
        "phi",
        "iter",
        "startp",
        "endp",
        "pixel",
        "width",
        "delta2theta",
        "device",
        "outdirs",
        "primary_outdir",
        "outprefix",
        "unscramble_method",
        "retry_timeout",
    ],
    output_names=["outfile"],
):
    def run(self):
        outdirs = self.get_input_value("outdirs", dict())
        primary_outdir = self.get_input_value("primary_outdir", None)
        outdirs = dirutils.prepare_outdirs(outdirs, primary_outdir)

        options = {
            "pars": self.inputs.parsfile,
            "debug": self.get_input_value("debug", False),
            "wavelength": self.get_input_value("wavelength", None),
            "energy": self.get_input_value("energy", None),
            "step": self.get_input_value("step", None),
            "range": self.get_input_value("range", [float("nan"), float("nan")]),
            "phi": self.get_input_value("phi", None),
            "iter": self.get_input_value("iter", None),
            "startp": self.get_input_value("startp", None),
            "endp": self.get_input_value("endp", None),
            "pixel": self.get_input_value("pixel", None),
            "width": self.get_input_value("width", None),
            "delta2theta": self.get_input_value("delta2theta", 0.0),
            "device": self.get_input_value("device", None),
        }

        sdelta2theta = str(options["delta2theta"]).replace(".", "")
        outfilename = (
            os.path.splitext(os.path.basename(self.inputs.filename))[0]
            + f"_w{sdelta2theta}.h5"
        )
        if not self.missing_inputs.outprefix:
            outfilename = self.inputs.outprefix + "_" + outfilename
        outfile = dirutils.primary_file(outfilename, outdirs)
        options["output"] = outfile

        if options["device"] is not None:
            options["device"] = str(options["device"])
        cl_platforms = pyopencl.get_platforms()
        for i, p in enumerate(cl_platforms):
            for j, d in enumerate(p.get_devices()):
                print(f"{i},{j}: {d}")

        filename = self.inputs.filename
        entries = self.get_input_value("entries", None)
        unscramble_method = self.get_input_value("unscramble_method", "xy")
        if True:
            print(f"Unscrambling of ROI collection: {unscramble_method}")
            retry_timeout = self.get_input_value("retry_timeout", 10)
            if os.path.exists(outfile):
                exclude_entries = safe_top_level_names(
                    outfile, retry_timeout=retry_timeout
                )
            else:
                exclude_entries = None
            options["hdf5_data"] = read_data(
                filename,
                unscramble_method=unscramble_method,
                entries=entries,
                exclude_entries=exclude_entries,
                retry_timeout=retry_timeout,
            )
        else:
            # Let multianalyzer read the data
            options["filename"] = filename
            options["entries"] = entries

        rebin_file(**options)

        dirutils.copy_file(outfilename, outdirs)

        self.outputs.outfile = outfile



from ewokscore import Task
from silx.io.h5py_utils import retry, File, safe_top_level_names
from silx.utils.retry import RetryError


@retry(retry_period=0.5)
def wait_scans_finished(filename, entries=None):
    if not entries:
        entries = safe_top_level_names(filename)
    with File(filename, mode="r") as fh:
        for entry in entries:
            if "end_time" not in fh[entry]:
                raise RetryError
            print(f"Scan {entry} is finished")


class WaitScansFinished(
    Task,
    input_names=["filename"],
    optional_input_names=[
        "entries",
        "retry_timeout",
    ],
    output_names=["filename", "entries"],
):
    def run(self):
        if self.inputs.entries:
            retry_timeout = self.get_input_value("retry_timeout", 10)
            print(f"Wait: {self.inputs.filename} ({self.inputs.entries})")
            wait_scans_finished(
                self.inputs.filename,
                entries=self.inputs.entries,
                retry_timeout=retry_timeout,
            )
        self.outputs.filename = self.inputs.filename
        self.outputs.entries = self.get_input_value("entries", None)



from pathlib import Path

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources


def program_path(program: str) -> Path:
    with resources.path(__name__, program) as path:
        return path


def main(argv=None):
    import sys
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Full path of the binary")
    parser.add_argument("name", type=str, help="Binary name")

    args = parser.parse_args(argv[1:])
    print(program_path(args.name))



import subprocess
import pytest
from ..bin import program_path


@pytest.mark.parametrize("program", ["id22sume", "id22sumalle"])
def test_binary(program):
    cmd = program_path(program)
    subprocess.check_output([str(cmd)])


@pytest.mark.parametrize("program", ["id22sumepy"])
def test_console_script(program):
    subprocess.check_output([program, "--help"])



import numpy
import random
from .. import roitools


def test_unscramble_roi_collection():
    x = [1, 2, 3]
    y = [4, 5, 6, 7]
    n = len(x) * len(y)
    mot = numpy.random.random(n)
    motfixed = numpy.random.random(n + 1)
    mca = numpy.random.random((3, n))
    mcafixed = numpy.random.random((3, n + 1))
    scalarfixed = 10

    data1 = {
        "x": numpy.repeat(x, len(y)),
        "y": numpy.tile(y, len(x)),
        "mot": mot,
        "motfixed": motfixed,
        "mca": mca,
        "mcafixed": mcafixed,
        "scalarfixed": scalarfixed,
    }

    idx = list(range(n))
    random.shuffle(idx)
    data2 = {k: v[..., idx] if "fixed" not in k else v for k, v in data1.items()}

    data3 = roitools.unscramble_roi_collection(data2, ["x", "y"])

    data1 = {k: v.tolist() if "scalar" not in k else v for k, v in data1.items()}
    data3 = {k: v.tolist() if "scalar" not in k else v for k, v in data3.items()}
    assert data1 == data3



import os
import shutil
import subprocess
from ewokscore import Task
from . import topas_extract
from . import dirutils


class ID22TopasExtract(
    Task,
    input_names=["filename", "inp_file"],
    optional_input_names=[
        "entries",
        "tth_min",
        "tth_max",
        "full_tth",
        "outdirs",
        "outprefix",
        "primary_outdir",
        "inp_step",
        "startp",
    ],
):
    def run(self):
        in_file = self.get_input_value("filename", None)

        scans = self.get_input_value("entries", None)
        if scans:
            scans = sorted({int(float(s)) for s in scans})
        assert len(scans) == 1, f"Extract only 1 scan ({scans})"
        tth_min = self.get_input_value("tth_min", None)
        tth_max = self.get_input_value("tth_max", None)
        full_tth = self.get_input_value("full_tth", False)

        outdirs = self.get_input_value("outdirs", dict())
        primary_outdir = self.get_input_value("primary_outdir", None)
        outdirs = dirutils.prepare_outdirs(outdirs, primary_outdir)

        out_file = os.path.splitext(os.path.basename(in_file))[0]
        if not self.missing_inputs.outprefix:
            out_file = self.inputs.outprefix + "_" + out_file
        out_file = out_file + "_calib"

        # TODO: use all outdirs
        rootpath = os.path.join(outdirs["primary"][1], out_file)
        savepath_xye = os.path.join(rootpath, "topas_extract")
        savepath_inp = os.path.join(rootpath, "topas_refine")
        os.makedirs(savepath_xye, exist_ok=True)
        os.makedirs(savepath_inp, exist_ok=True)

        topas_extract.run(
            in_file,
            out_file,
            scans=scans,
            savepath=savepath_xye,
            min_usr_tth=tth_min,
            max_usr_tth=tth_max,
            full_tth=full_tth,
        )

        scannr = scans[0]
        savepath_xye = os.path.relpath(savepath_xye, savepath_inp)
        xye_files = os.path.join(savepath_xye, out_file + f"_s{scannr}_MA")
        inp_file = self.inputs.inp_file
        inp_step = self.get_input_value("inp_step", 2)
        startp = self.get_input_value("startp", 31)

        inp_file_cp = os.path.join(savepath_inp, os.path.basename(inp_file))
        shutil.copyfile(inp_file, inp_file_cp)

        # Modify inp_file_cp and save xdds.inp
        args = [
            "gen_xdds_file",
            xye_files,
            os.path.basename(inp_file_cp),
            0,
            topas_extract.NUM_CHANNELS - 1,
            startp,
            topas_extract.X_HIGH,
            inp_step,
        ]
        args = list(map(str, args))

        proc = subprocess.Popen(
            args, stdout=subprocess.PIPE, cwd=os.path.dirname(inp_file_cp)
        )
        try:
            outs, errs = proc.communicate(timeout=120)
        except subprocess.TimeoutExpired:
            proc.kill()
            raise
        with open(os.path.join(savepath_inp, "xdds.inp"), "wb") as f:
            f.write(outs)



"""
Written by Ola G. Grendal @ ID22 ESRF
"""

import os
import sys
import numpy as np
import h5py
import bisect
from concurrent.futures import ProcessPoolExecutor as Pool


X_LOW = 0
X_HIGH = 511
NUM_CHANNELS = 13
ROIS = [i for i in range(NUM_CHANNELS)]
ROWS = [i for i in range(X_LOW, X_HIGH + 1)]

# TODO check for sorted/non-unique tth. !!Chop data fixed when data is not unique!!


def minmax_tth(min_usr_tth, max_usr_tth, tth_min, tth_max):
    # Takes in the user set min/max tth, and min/max of scans, and return min/max tth for data to to exported to xye-files
    print(f"tth min - max is: {tth_min} - {tth_max}")
    print(
        f"Widest range possible for MA06 with all detectors: {tth_min+12} - {tth_max-12}"
    )
    print(f"Your selection for MA06 is: {min_usr_tth} - {max_usr_tth}")

    if min_usr_tth is None and max_usr_tth is None:
        print(
            f"Your MA06 range is set to the widest possible: {tth_min+12}-{tth_max-12}"
        )
        return tth_min + 12, tth_max - 12
    elif min_usr_tth is None and max_usr_tth is not None:
        print("Your MA06 range is {tth_min+12} - {min(max_usr_tth, tth_max)}")
        return tth_min + 12, min(max_usr_tth, tth_max)
    elif min_usr_tth is not None and max_usr_tth is None:
        print(f"Your MA06 range is: {max(min_usr_tth, tth_min)} - {tth_max-12}")
        return max(min_usr_tth, tth_min), tth_max - 12
    else:
        print(
            f"Your MA06 range is: {max(min_usr_tth, tth_min)} - {min(max_usr_tth, tth_max)}"
        )
        return max(min_usr_tth, tth_min), min(max_usr_tth, tth_max)


def set_savepath(savepath):
    # Returns a string with the path to where to save the output images
    if savepath is None:
        return str("./")
    else:
        os.makedirs(savepath, exist_ok=True)
        return savepath


def set_filename(scan, in_file, out_file):
    # Returns a list of filename-strings
    if out_file is None:
        file_name_base = str(in_file.split("/")[-1].split(".")[0])
    else:
        file_name_base = out_file
    out_file_name_base = file_name_base + "_s{}_".format(scan)
    return out_file_name_base


def read_h5(scan, in_file):
    # Read h5-file and returns dict with info
    print(f"Reading data from scan {scan} of {in_file}...")
    res = {}
    with h5py.File(in_file, "r") as h:
        res["roicol"] = h["/{}.1/measurement/eiger_roi_collection".format(scan)][()]
        res["tth"] = h["/{}.1/measurement/tth".format(scan)][()]
        res["mon"] = h["/{}.1/measurement/mon".format(scan)][()]
        res["xs"] = h["/{}.1/instrument/eiger_roi_collection/selection/x".format(scan)][
            ()
        ]  # FIX
        res["ys"] = h["/{}.1/instrument/eiger_roi_collection/selection/y".format(scan)][
            ()
        ]  # FIX
    return res


def get_tth_indexes(tth_minMA06, tth_maxMA06, tth):
    # Find the tth subranges for the detectors assuming a nominal offset of 2-deg, and the corresponding indexes
    sub_ranges_tth = {}
    sub_ranges_idx = {}
    for i in range(NUM_CHANNELS):
        tth_low = tth_minMA06 + (6 - i) * 2
        tth_high = tth_maxMA06 + (6 - i) * 2
        idx_low = bisect.bisect_left(tth, tth_low)
        idx_high = bisect.bisect_left(tth, tth_high)
        sub_ranges_tth[f"MA{i:02d}"] = [tth_low, tth_high]
        sub_ranges_idx[f"MA{i:02d}"] = [idx_low, idx_high]
    # print(sub_ranges_tth, sub_ranges_idx)
    return sub_ranges_idx


def set_scans(scans, ex_scans):
    if isinstance(scans, int):
        all_scans = [scans]
    elif len(scans) == 1:
        all_scans = scans
    elif len(scans) == 2:
        all_scans = [x for x in range(scans[0], scans[1] + 1)]
        if ex_scans is not None:
            all_scans = [x for x in all_scans if x not in ex_scans]
    elif len(scans) >= 3:
        all_scans = [x for x in scans]
        if ex_scans is not None:
            all_scans = [x for x in all_scans if x not in ex_scans]
    else:
        all_scans = tuple()
    assert all_scans, "no scans to process"
    print("Converting the following scan(s): ", end="")
    print(*all_scans, sep=", ")
    return all_scans


# def check_tth(tth):
# 	is_sorted = np.all(tth[:-1] <= tth[1:])
# 	if is_sorted is False:
# 		print('!!tth is not a sorted array!!')
# 	tth_set = set(tth)
# 	if len(tth_set) == len(tth):
# 		all_unique = True
# 		chop = None
# 	else:
# 		all_unique = False
# 		chop = len(tth) - len(tth_set)
# 		print('!!tth does not only contain unique values!!')
# 	return is_sorted, all_unique, chop


def run(
    in_file,
    out_file,
    scans=(1,),
    savepath=None,
    ex_scans=None,
    min_usr_tth=None,
    max_usr_tth=None,
    full_tth=False,
):
    # Saves data from roi_collection to .xye-files
    save_data_path = set_savepath(savepath)
    all_scans = set_scans(scans, ex_scans)
    fmt = ("%3.8f", "%i", "%i")
    for scan in all_scans:
        data = read_h5(scan, in_file)
        out_filename = set_filename(scan, in_file, out_file)

        # is_sorted, all_unique, chop = check_tth(data['tth'])
        # if all_unique is False:
        # 	data['tth'] = data['tth'][chop:]
        # 	data['mon'] = data['mon'][chop:]
        # 	data['roicol'] = data['roicol'][chop:]

        sorted_xs = sorted(set(data["xs"]))
        nfiles = len(data["xs"])

        print(f"Extract scan {scan} from file {in_file} ...")

        if full_tth:
            with Pool() as pool:
                for i in range(nfiles):
                    channel = sorted_xs.index(data["xs"][i])
                    suffix = "MA{}_c{}.xye".format(channel, data["ys"][i])
                    filenamei = os.path.join(save_data_path, out_filename + suffix)
                    datai = np.c_[data["tth"], data["roicol"][:, i], data["mon"]]
                    pool.submit(np.savetxt, filenamei, datai, fmt=fmt)
        else:
            tth_min = round(data["tth"][0])
            tth_max = round(data["tth"][-1])
            tth_minMA06, tth_maxMA06 = minmax_tth(
                min_usr_tth, max_usr_tth, tth_min, tth_max
            )
            ranges_idx = get_tth_indexes(tth_minMA06, tth_maxMA06, data["tth"])

            with Pool() as pool:
                for i in range(nfiles):
                    channel = sorted_xs.index(data["xs"][i])
                    low_idx = ranges_idx[f"MA{channel:02d}"][0]
                    high_idx = ranges_idx[f"MA{channel:02d}"][1]
                    suffix = "MA{}_c{}.xye".format(channel, data["ys"][i])
                    filenamei = os.path.join(save_data_path, out_filename + suffix)
                    datai = np.c_[
                        data["tth"][low_idx:high_idx],
                        data["roicol"][low_idx:high_idx, i],
                        data["mon"][low_idx:high_idx],
                    ]
                    pool.submit(np.savetxt, filenamei, datai, fmt=("%3.8f", "%i", "%i"))

        print(f"Saved {nfiles} xye files")


def main(argv=None):
    import argparse
    from datetime import datetime

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Script for exporting xye-files for TOPAS from roi collection @ ID22, ESRF"
    )
    parser.add_argument(
        "-in",
        "--in_file",
        type=str,
        required=True,
        help="Filepath to h5-file including filename",
    )

    parser.add_argument(
        "-s",
        "--scans",
        type=int,
        default=1,
        nargs="+",
        help="Scan number(s) of scan(s) to convert. Default is scan 1.1. Ex: 2 or 1 10 or 2 3 5",
    )

    parser.add_argument(
        "-es",
        "--ex_scans",
        type=int,
        default=None,
        nargs="+",
        help="Scan number(s) of scan(s) to exclude. Default is none. Ex: 2 or 2 3 5",
    )

    parser.add_argument(
        "-l_tth",
        "--tth_min",
        type=float,
        default=None,
        help="Lowest tth to include (MA6). Default is lowest possible (scan_start + 12).",
    )
    parser.add_argument(
        "-h_tth",
        "--tth_max",
        type=float,
        default=None,
        help="Highest tth to include (MA6). Default is highest possible (scan_end - 12).",
    )
    parser.add_argument(
        "-full_tth",
        "--full_tth",
        type=int,
        default=0,
        help="Set 1 to use all data/full tth-range. Default is 0.",
    )

    parser.add_argument(
        "-o",
        "--savepath",
        type=str,
        default=None,
        help="Full path to where you want to save your data. By default data is saved in the current directory: ./",
    )

    parser.add_argument(
        "-f",
        "--out_file",
        type=str,
        default=None,
        help="Filename. Ex: LaB6_35keV --> LaB6_35keV_sx_MAy_z.xye. If not given filename of h5-file is used --> h5filename_sx_MAy_z.xye.",
    )

    # parser.add_argument('-is', '--includescan', type=int, nargs='+', default=None, help='If used alone list of the only scans to be included. If used with -sts and/or -ens it will include scans outside the given range. Ex: 2 3 5 10')

    args = parser.parse_args(argv[1:])

    time_start_TOT = datetime.now()
    run(
        args.in_file,
        args.out_file,
        scans=args.scans,
        savepath=args.savepath,
        ex_scans=args.ex_scans,
        min_usr_tth=args.tth_min,
        max_usr_tth=args.tth_max,
        full_tth=bool(args.full_tth),
    )
    print("Time total: {}".format(datetime.now() - time_start_TOT))


if __name__ == "__main__":
    sys.exit(main())



import os


def saved_scan_numbers(filename):
    """Scans saved in the SPEC file.

    :param str filename:
    :param dict outdirs:
    :returns list(int):
    """
    saved = []
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            for line in f:
                if "#S" in line:
                    saved.append(int(line.split()[1]))
    return saved



"""Convert an HDF5 file to a SPEC .dat file
"""

import os
import io
import datetime
import logging

import numpy
from scipy.interpolate import interp1d
from silx.io.h5py_utils import retry, File

from . import dirutils
from . import specutils

logger = logging.getLogger(__name__)


def convert_h5(
    raw_filename,
    outprefix=None,
    entries=None,
    outdirs=None,
    primary_outdir=None,
    retry_timeout=10,
    rebin_filename=None,
    ascii_extension=".dat",
):
    """
    :param str raw_filename: full path of the Nexus file
    :param str outprefix: something unique to the proposal/session
    :param list entries: for example ["1.1", "1.2", ...]
    :param dict outdirs:
    :param str primary_outdir:
    :param str rebin_filename: full path of the id22rebin file
    :returns str: primary spec output file
    """
    raw_filename = os.path.abspath(raw_filename)
    outdirs = dirutils.prepare_outdirs(outdirs, primary_outdir)
    if "primary" not in outdirs:
        logger.warning("No primary output directory: not saving anything")
        return

    if entries:
        names = {}
        for name in entries:
            scannr = int(name.split(".")[0])
            names.setdefault(scannr, []).append(name)
    else:
        names = get_scan_names(raw_filename, retry_timeout=retry_timeout)
    if not names:
        logger.warning("no scans to convert")
        return

    # Save SPEC header when no scans have been saved yet
    if rebin_filename:
        spec_filename = rebin_filename
    else:
        spec_filename = raw_filename
    spec_filename = (
        os.path.splitext(os.path.basename(spec_filename))[0] + ascii_extension
    )
    if outprefix:
        spec_filename = outprefix + "_" + spec_filename
    has_new_data = False

    saved = saved_scan_numbers(spec_filename, outdirs)
    if not saved:
        first_scan = sorted(names.items())[0][1][0]
        start_time = get_start_time(
            raw_filename, first_scan, retry_timeout=retry_timeout
        )
        specdata = create_spec_header(raw_filename, start_time=start_time)
        has_new_data |= bool(specdata)
        add_to_specfile(spec_filename, specdata, outdirs)

    # Save scans
    first_error = None
    for scannr, subscans in sorted(names.items()):
        if scannr in saved:
            # fscan already saved
            continue
        if len(subscans) != 2:
            # incomplete fscan
            continue
        # fscan with 2 complete subscans
        subscan1, subscan2 = subscans
        try:
            specdata = read_fscan_data(
                raw_filename,
                subscan1,
                subscan2,
                rebin_filename=rebin_filename,
                retry_timeout=retry_timeout,
            )
        except Exception as e:
            if first_error is None:
                first_error = e
            continue
        has_new_data |= bool(specdata)
        add_to_specfile(spec_filename, specdata, outdirs)

    if has_new_data:
        dirutils.copy_file(spec_filename, outdirs)

    if first_error is not None:
        raise first_error

    return dirutils.primary_file(spec_filename, outdirs)


@retry(retry_period=0.5, retry_timeout=10)
def get_scan_names(filename, title=None):
    """Get the subscan names for all scans in the Nexus file

    :param str filename:
    :param str title:id22
    :returns dict: scannr(int)->subscan_names(list)
    """
    with File(filename, mode="r") as h5file:
        names = list(h5file["/"])

        def include(name):
            try:
                scan = h5file[name]
            except Exception as e:
                logger.warning(
                    "cannot read scan " + repr(name) + " (cause: " + str(e) + ")"
                )
                return False
            if "end_time" not in scan:
                return False
            if "measurement" not in scan:
                return False
            if title:
                stitle = str_from_dataset(scan["title"])
                if not any(s in stitle for s in ["fscan", "f2scan"]):
                    return False
            return True

        scans = dict()
        for name in names:
            if include(name):
                scannr = int(float(name))
                scans.setdefault(scannr, []).append(name)

        return scans


def saved_scan_numbers(filename, outdirs):
    """Scans saved in the SPEC file.

    :param str filename:
    :param dict outdirs:
    :returns list(int):
    """
    local_filename = dirutils.primary_file(filename, outdirs)
    return specutils.saved_scan_numbers(local_filename)


def add_to_specfile(spec_filename, specdata, outdirs):
    """
    :param str spec_filename:
    :param list(2-tuple) specdata:
    :param dict outdirs:
    """
    if not outdirs:
        return
    local_filename = dirutils.primary_file(spec_filename, outdirs)
    dirname = os.path.dirname(local_filename)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    for mode, lines in specdata:
        with open(local_filename, mode) as f:
            f.writelines(lines)


MOTOR_NAMES = [
    ["tth", "om", "manom", "mantth", "mantr", "Dh", "Dhd"],
    ["Dhm", "Dhu", "spinp", "bluspin", "t1h", "t1h1", "t1h2", "t1x"],
    ["t1y", "t1rz", "t1trans", "robtran", "s3vg", "s3vo", "s3hg", "s3ho"],
    ["s4vg", "s4vo", "s4hg", "s4ho", "u26b", "chi"],
    ["d2dtran", "Dy", "Dyu", "Dyd", "Drx", "Dry"],
    ["mos", "rst", "rsg", "redtrans"],
    ["xtrans", "ytrans", "ztrans", "gasspin", "DET_Z", "DET_X", "DET_Y", "DET_RZ"],
]


def create_spec_header(filename, start_time=None):
    """
    :param str filename:
    :returns list(2-tuple):
    """
    specdata = []
    lines = []
    specdata.append(("w", lines))

    if not start_time:
        start_time = datetime.datetime.now().astimezone().isoformat()
    lines.append("#F " + '"' + filename + '"' + "\n")
    lines.append("#D " + start_time + "\n")
    lines.append("#C exp  User = opid22" + "\n")
    lines.append("" + "\n")
    for i, names in enumerate(MOTOR_NAMES):
        lines.append("#o{} ".format(i) + " ".join(names) + "\n")
    lines.append("\n")
    return specdata


@retry(retry_period=0.5, retry_timeout=10)
def get_start_time(filename, scan):
    """
    :param str filename:
    :param str scan:
    :returns str:
    """
    with File(filename, mode="r") as h5file:
        return str_from_dataset(h5file[scan]["start_time"])


@retry(retry_period=0.5, retry_timeout=10)
def read_fscan_data(raw_filename, subscan1, subscan2, rebin_filename=None):
    """
    :param str raw_filename:
    :param str subscan1:
    :param str subscan2:
    :param str rebin_filename:
    :returns list(2-tuple):
    """
    with File(raw_filename, mode="r") as h5file:
        gsubscan1 = h5file[subscan1]
        gsubscan2 = h5file[subscan2]
        if rebin_filename:
            with File(rebin_filename, mode="r") as h5filerebin:
                try:
                    rebinscan = h5filerebin[subscan1]
                except KeyError:
                    return list()
                return _read_fscan_data(gsubscan1, gsubscan2, rebinscan=rebinscan)
        else:
            return _read_fscan_data(gsubscan1, gsubscan2)


def _read_fscan_data(subscan1, subscan2, rebinscan=None):
    """
    :param h5py.Group subscan1:
    :param h5py.Group subscan2:
    :param h5py.Group rebinscan:
    :returns list(2-tuple):
    """
    specdata = []

    fast_data = subscan1["measurement"]
    slow_data = subscan2["measurement"]
    if rebinscan is None:
        rebin_data = None
    else:
        rebin_data = rebinscan["id22rebin/data"]

    fscan_params = subscan1["instrument/fscan_parameters"]
    positioners_start = subscan1["instrument/positioners_start"]
    try:
        machine = subscan1["instrument/machine"]
    except KeyError:
        machine = {}
    try:
        robot = subscan1["instrument/robot"]
    except KeyError:
        robot = {}

    # Scan parameters
    scannr = subscan1.name[1:].split(".")[0]
    start_time = str_from_dataset(subscan1["start_time"])
    start_pos = float(fscan_params["start_pos"][()])
    step = float(fscan_params["step_size"][()])
    no_scan_points = float(fscan_params["npoints"][()])
    acq_time = float(fscan_params["acq_time"][()])
    end_pos = "{:.2f}".format(start_pos + step * no_scan_points)
    deg_per_min = "{:.2f}".format(step / acq_time * 60)

    # Scan header
    lines = []
    specdata.append(("a", lines))

    lines.append(
        "#S  "
        + scannr
        + "  hookscan "
        + str_from_dataset(fscan_params["motor"])
        + " "
        + read_position(fscan_params, "start_pos", "{:.2f}")
        + " "
        + end_pos
        + " "
        + deg_per_min
        + " "
        + read_position(fscan_params, "acq_time", "{:.5f}", modif=lambda x: x * 1000)
        + " "
        + "\n"
    )
    lines.append("#D " + start_time + "\n")
    lines.append(
        "#T " + read_position(fscan_params, "acq_time", "{:.5f}") + " (Seconds)\n"
    )
    lines.append("#Q \n")

    for i, names in enumerate(MOTOR_NAMES):
        positions = " ".join(
            [read_position(positioners_start, name, "{:.4f}") for name in names]
        )
        lines.append("#P{} ".format(i) + positions + "\n")

    lines.append("#UMI0    Current     AutoM      Shutter      U26B_GAP     \n")
    lines.append(
        "#UMI1"
        + " "
        + read_position(machine, "current", "{:.4f}")
        + " "
        + str_from_dataset(machine.get("automatic_mode"))
        + " "
        + str_from_dataset(machine.get("front_end"))
        + " "
        + read_position(positioners_start, "u26b", "{:.4f}")
        + "\n"
    )

    lines.append(
        "#UMI2"
        + " Refill in "
        + str_from_dataset(machine.get("refill_countdown"))
        + " sec,"
        + " Fill Mode: "
        + str_from_dataset(machine.get("mode"))
        + ","
        + " Op. Message: "
        + str_from_dataset(machine.get("message"))
        + "\n"
    )

    lines.append(
        "#CR"
        + " Last robot sample loaded: "
        + str_from_dataset(robot.get("sample_label"))
        + "\n"
    )

    # Slow counters
    slow_ctrs_spec = [
        "blowerT",
        "Cryostream",
        "Cryostat",
        "Press_in",
        "Press_out",
        "monin",
        "bmon",
    ]
    slow_ctrs_fmt = ["%4.3f", "%4.3f", "%4.3f", "%7.4f", "%7.4f", "%.5e", "%.5e"]
    slow_ctrs_h5 = [
        ("blower_in", False),
        ("ox700", False),
        ("ls340_A", False),
        ("pace_in", False),
        ("pace_press", False),
        ("monin", False),
        ("bmon", False),
    ]
    nslow_ctrs = len(slow_ctrs_spec)
    npts_slow = min_npts_ctrs(slow_data, slow_ctrs_h5)

    # Fast counters
    if "eiger" in fast_data:
        prefix = "eiger_roi"
        nchannels = 13
    else:
        prefix = "ma"
        nchannels = 9
    if rebin_data is None:
        fast_ctrs_spec = (
            ["2_theta"]
            + ["MA{}".format(i) for i in range(nchannels)]
            + ["Monitor", "Epoch", "Omega"]
        )
        fast_ctrs_fmt = ["%3.8f"] + ["%i"] * nchannels + ["%i", "%15.8f", "%3.8f"]
        fast_ctrs_h5 = (
            ["tth"]
            + [prefix + str(i) for i in range(nchannels)]
            + ["mon", "epoch_trig", "om"]
        )
    else:
        fast_ctrs_spec = ["Epoch", "Omega"]
        fast_ctrs_fmt = ["%3.8f"] * len(fast_ctrs_spec)
        fast_ctrs_h5 = ["epoch_trig", "om"]
    not_required = ["om"]
    fast_ctrs_h5 = [(name, name not in not_required) for name in fast_ctrs_h5]
    nfast_ctrs = len(fast_ctrs_spec)
    npts_fast = min_npts_ctrs(fast_data, fast_ctrs_h5)

    # Rebin counters
    if rebin_data is None:
        rebin_ctrs_h5 = []
        rebin_ctrs_spec = []
        rebin_ctrs_fmt = []
        rebin_ctrs_h5 = []
    else:
        rebin_ctrs_h5 = ["2th", "I_sum", "norm"]
        rebin_ctrs_spec = ["2_theta"] + [
            f"{name}{i}"
            for name in ["MA", "Mon"]
            for i in list(range(nchannels)) + ["av"]
        ]
        rebin_ctrs_fmt = (
            ["%3.8f"] + ["%i"] * nchannels + ["%3.8f"] + ["%i"] * nchannels + ["%3.8f"]
        )
        rebin_ctrs_h5 = [(name, True) for name in rebin_ctrs_h5]
    nrebin_ctrs = len(rebin_ctrs_spec)
    npts_rebin = min_npts_ctrs(rebin_data, rebin_ctrs_h5)

    # Prepare data
    if npts_rebin:
        nrows = npts_rebin
    else:
        nrows = npts_fast
    ncols = nrebin_ctrs + nfast_ctrs + nslow_ctrs
    data = numpy.zeros((nrows, ncols))
    ctrs_spec = rebin_ctrs_spec + fast_ctrs_spec + slow_ctrs_spec
    ctrs_fmt = rebin_ctrs_fmt + fast_ctrs_fmt + slow_ctrs_fmt
    rebinoff = 0
    fastoff = nrebin_ctrs
    slowoff = nrebin_ctrs + nfast_ctrs

    # Read rebin data
    off = rebinoff
    for i, idata in read_ctrs(rebin_data, rebin_ctrs_h5, npts_rebin):
        if idata.ndim == 2:
            idata = idata.T
            nadd = idata.shape[-1]
            data[:, off : off + nadd] = idata
            off += nadd
            data[:, off] = numpy.mean(idata, axis=1)
            off += 1
        else:
            data[:, off] = idata
            off += 1

    # Read fast data + interpolate at rebinned 2-theta
    if npts_rebin:
        xnew = list(read_ctrs(rebin_data, [("2th", True)], npts_rebin))[0][-1]
        xold = list(read_ctrs(fast_data, [("tth", True)], npts_fast))[0][-1]
        for i, idata in read_ctrs(fast_data, fast_ctrs_h5, npts_fast):
            func = interp1d(xold, idata, kind="nearest", fill_value="extrapolate")
            try:
                data[:, fastoff + i] = func(xnew)
            except Exception:
                pass
    else:
        for i, idata in read_ctrs(fast_data, fast_ctrs_h5, npts_fast):
            data[:, fastoff + i] = idata

    # Read slow data + interpolate at fast epoch
    xold = list(read_ctrs(slow_data, [("epoch", True)], npts_slow))[0][-1]
    xnew = data[:, ctrs_spec.index("Epoch")]
    for i, idata in read_ctrs(slow_data, slow_ctrs_h5, npts_slow):
        func = interp1d(xold, idata, kind="nearest", fill_value="extrapolate")
        try:
            data[:, slowoff + i] = func(xnew)
        except Exception:
            pass

    # Scan data header
    lines = []
    specdata.append(("a", lines))
    lines.append("#N {}\n".format(ncols))
    lines.append("#L  " + "  ".join(ctrs_spec) + "\n")

    # Scan data
    lines = []
    specdata.append(("ab", lines))
    f = io.BytesIO()
    numpy.savetxt(f, data, delimiter=" ", fmt=" ".join(ctrs_fmt))
    lines.append(f.getbuffer())
    lines.append(b"\n")

    return specdata


def read_ctrs(group, ctrs, npts):
    """Read datasets

    :param h5py.Group group:
    :param list(2-tuple) ctrs:
    :param int npts:
    :yield numpy.ndarray:
    """
    for i, (name, must_exist) in enumerate(ctrs):
        try:
            dset = group[name]
        except KeyError:
            if must_exist:
                raise
        else:
            try:
                data = dset[:npts]
            except Exception as e:
                logger.warning(
                    "skip counter data " + repr(name) + " (cause: " + str(e) + ")"
                )
            else:
                if not len(data):
                    logger.warning("no data in " + repr(name))
                    continue
                yield i, data


def min_npts_ctrs(group, ctrs):
    """Smallest number of points of a group of datasets.

    :param h5py.Group group:
    :param list(2-tuple) ctrs:
    :returns int:
    """
    if not ctrs:
        return 0
    npts = []
    for name, must_exist in ctrs:
        try:
            dset = group[name]
        except KeyError:
            if must_exist:
                raise
        else:
            npts.append(dset.shape[-1])
    return min(npts)


def str_from_dataset(dataset):
    """Read dataset as a string

    :param h5py.Dataset dataset:
    :returns str:
    """
    if isinstance(dataset, str):
        return dataset
    if dataset is None:
        return "UNKNOWN"
    try:
        return dataset.asstr()[()]
    except (AttributeError, TypeError):
        return str(dataset[()])


def read_position(grp, key, fmt, modif=None):
    """Read a motor position from grp[key], return "-999" when missing.

    :param h5py.Group grp:
    :param str key:
    :param callable or None modif:
    :returns str:
    """
    if key in grp:
        pos = grp[key][()]
        if pos == "*DIS*":
            return str(-999)
        try:
            num = float(fmt.format(pos))
        except Exception as e:
            raise RuntimeError("Error in formatting motor position " + repr(key)) from e
        if modif:
            num = modif(num)
        return str(num)
    return str(-999)



"""Script for batch runs with id22sum and id22sumall:
processes powder diffraction data from a multianalyser instrument

Fortran source code: https://github.com/jonwright/id31sum

Written by Ola G. Grendal @ ID22 ESRF
"""

import os
import sys
import subprocess
import glob

import numpy
import h5py
import hdf5plugin  # noqa F401

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None


from . import specutils


def usage_msg(name=None):
    """Function for re-defining the usage-message when no and/or wrong input is given"""

    return """id22sumepy

[-h, for more help than given here]
    -i, start of, or full filename. Ex: ch5900_S1_ or ch5900_S1_RT_0001.dat
    -b, binsize. Ex: 0.002
[-int, True for interactive mode, default is False]
"""


def epilog_msg(name=None):
    """Function for re-defining the usage-message when no and/or wrong input is given"""
    return """
PRO TIP #1):
Put filename in quotes with * or ? for linux wildcard expansion.
Ex: -i \"hc4000_sample*_\"
"""


def generate_cmd(
    program,
    programdir,
    filename,
    binsize,
    first_scan,
    last_scan,
    lowtth,
    scaling,
    excludedetector,
    excludescan,
    zingit,
    advanced=None,
):
    """Fortran sum program command"""
    if programdir:
        program = os.path.join(programdir, program)
    cmd = [
        program,
        filename,
        binsize,
        "{}".format(first_scan),
        "{}".format(last_scan),
        "lowtth={}".format(lowtth),
    ]
    if scaling is None:
        cmd.append("scalmon")
    elif scaling.startswith("scal"):
        cmd.append("{}".format(scaling))
    if excludescan:
        cmd.append("es={}".format(excludescan))
    if excludedetector:
        cmd.append("ed={}".format(excludedetector))
    if zingit:
        cmd.append("ef={}.zin".format(os.path.basename(filename)))
    if advanced is not None:
        for i in advanced:
            cmd.append(i)
    return cmd


def generate_es(first_scan, last_scan, *args):
    """Exclude scan numbers: exclude from start, missing from spec file, interactive prompt"""
    scannrs = list()
    for numbers in args:
        if numbers:
            if isinstance(numbers, int):
                scannrs.append(numbers)
                continue
            if isinstance(numbers, str):
                numbers = numbers.split(",")
            scannrs += [int(i) for i in numbers]
    scannrs = [i for i in sorted(set(scannrs)) if i >= first_scan and i <= last_scan]
    if scannrs:
        return ",".join(list(map(str, scannrs)))
    return None


def yes_no(question):
    """Yes/no prompt. Returns True or False."""
    answers = ["n", "no", "y", "yes"]
    while True:
        try:
            answer = input(question)
        except ValueError:
            print("Sorry, I did not understand that!")
            continue
        answer = answer.lower()
        if answer not in answers:
            print("Your response must be [y]es or [n]o")
            continue
        else:
            break
    if answer.startswith("y"):
        return True
    else:
        return False


def new_bin(question):
    """Retrieve new binsize"""
    while True:
        try:
            answer = float(input(question))
            break
        except ValueError:
            print("Please type a valid number. Ex: 0.002")
            continue

    return str(answer)


def removed_scans(question):
    """Retrieve scans to be removed"""
    while True:
        answer = input(question)
        try:
            scannrs = [int(i) for i in answer.split(",")]
            break
        except ValueError:
            print("Please type a valid number or list of numbers. Ex: 2 or 1,2,5")
    return ",".join(map(str, scannrs))


def plot_xy(
    filename,
    first_scan,
    last_scan,
    es_scans,
    sum_single,
    sum_all,
    binsize=None,
    show=True,
):
    """Reads XY data and plots it"""
    if es_scans is None:
        exclude_scans = []
    else:
        exclude_scans = [int(x) for x in es_scans.split(",")]
    all_scans = list(range(first_scan, last_scan + 1))
    plot_scans = [x for x in all_scans if x not in exclude_scans]
    if sum_all:
        for i in plot_scans:
            xyfilename = "{}_{}.xye".format(os.path.splitext(filename)[0], i)
            data = numpy.loadtxt(xyfilename, usecols=(0, 1)).T
            plt.plot(data[0], data[1], label="Scan: {}".format(i))

    if sum_single:
        if binsize:
            xyfilename = "{}_b{}.xye".format(
                os.path.splitext(filename)[0], str(binsize).replace(".", "")
            )
        else:
            xyfilename = "{}.xye".format(os.path.splitext(filename)[0])
        data = numpy.loadtxt(xyfilename, usecols=(0, 1)).T
        plt.plot(data[0], data[1], label="Final pattern")

    plt.plot(
        [], [], " ", label="To continue press: Ctrl + w"
    )  # Simple hack for adding explanatory text to legend
    plt.legend()
    plt.xlim(data[0].min(), data[0].max())
    plt.title(str(filename))
    if show:
        plt.show()


def find_h5_files(user_str):
    all_files = []
    if any(element in user_str for element in ["*", "?", "["]):
        all_files = glob.glob(user_str + ".h5")
    else:
        basename = os.path.basename(user_str)
        dirname = os.path.dirname(user_str)
        if not dirname:
            dirname = "."
        for filename in os.listdir(dirname):
            if filename.startswith(basename) and filename.endswith(".h5"):
                all_files.append(os.path.join(dirname, filename))
    print("id22sumpy: Found {} files starting with {}".format(len(all_files), user_str))
    return all_files


def get_all_scans_h5(file_path):
    with h5py.File(file_path, "r") as h:
        all_scans = sorted(h.keys(), key=lambda x: float(x))
    return all_scans


def read_h5(file_path, entry="1.1"):
    """Read h5-file and returns dict with data"""
    print("id22sumpy: Reading data...")
    res = {}
    with h5py.File(file_path, "r") as h:
        res["I"] = numpy.nan_to_num(
            h["/{}/id22rebin/data/I_MA".format(entry)][()], nan=0
        )
        res["I_sum"] = h["/{}/id22rebin/data/I_sum".format(entry)][()]
        res["tth"] = h["/{}/id22rebin/data/2th".format(entry)][()]
        res["norm"] = h["/{}/id22rebin/data/norm".format(entry)][()]
    return res


def save_spec(file_path, data, monav, entry="1.1"):
    scannr = int(float(entry))
    header_1 = "#S {} hookscan \n".format(scannr)
    header_2 = "#N 28\n"
    header_3 = "#L  2_theta  MA0  MA1  MA2  MA3  MA4  MA5  MA6  MA7  MA8  MA9  MA10  MA11  MA12  Mon0  Mon1  Mon2  Mon3  Mon4  Mon5  Mon6  Mon7  Mon8  Mon9  Mon10  Mon11  Mon12  Monav"
    header = header_1 + header_2 + header_3
    fmt = "%3.8f" + 13 * " %i" + 14 * " %i"
    if not os.path.exists(file_path):
        dirname = os.path.dirname(file_path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        with open(file_path, "wb+") as dat:
            numpy.savetxt(
                dat,
                numpy.c_[
                    data["tth"],
                    numpy.transpose(data["I_sum"]),
                    numpy.transpose(data["norm"]),
                    numpy.transpose(monav),
                ],
                header=header,
                comments="",
                fmt=fmt,
            )
    else:
        with open(file_path, "ab+") as dat:
            numpy.savetxt(
                dat,
                numpy.c_[
                    data["tth"],
                    numpy.transpose(data["I_sum"]),
                    numpy.transpose(data["norm"]),
                    numpy.transpose(monav),
                ],
                header="\n" + header,
                comments="",
                fmt=fmt,
            )
    print("id22sumpy: Saved scan {} to {}".format(scannr, file_path))
    return scannr


def execute_cmd(cmd, wd):
    cmd = list(map(str, cmd))
    print("\nid22sumpy: Execute " + " ".join(cmd))
    output = subprocess.check_output(cmd, universal_newlines=True, cwd=wd)
    if output:
        print(output)


def execute_sum(
    programdir,
    filename,
    binsize,
    first_scan,
    last_scan,
    lowtth,
    scaling,
    zingit,
    sum_single,
    sum_all,
    excludedetector,
    excludescan,
    outdir,
    advanced=None,
    interactive=0,
):
    """Execute id22sumalle and id22sume"""
    cmdfilename = os.path.abspath(filename)

    if sum_all:
        cmd = generate_cmd(
            "id22sumalle",
            programdir,
            cmdfilename,
            binsize,
            first_scan,
            last_scan,
            lowtth,
            scaling,
            excludedetector,
            excludescan,
            zingit,
            advanced=advanced,
        )
        execute_cmd(cmd, outdir)

    if sum_single:
        cmd = generate_cmd(
            "id22sume",
            programdir,
            cmdfilename,
            binsize,
            first_scan,
            last_scan,
            lowtth,
            scaling,
            excludedetector,
            excludescan,
            zingit,
            advanced=advanced,
        )
        execute_cmd(cmd, outdir)

    if interactive:
        plot_xy(filename, first_scan, last_scan, excludescan, sum_single, sum_all)


def run(
    filename,
    binsize,
    lowtth=0,
    scaling=None,
    entries=None,
    excludescan=None,
    excludedetector=None,
    interactive=0,
    sum_single=1,
    sum_all=1,
    zingit=0,
    advanced=tuple(),
    outdir=None,
    resbasename="temp.res",
    ascii_extension=".dat",
    programdir=None,
):
    if outdir is None:
        outdir = os.getcwd()
    resfile = os.path.join(outdir, resbasename)
    do_sum = sum_single or sum_all
    if do_sum and not os.path.isfile(resfile):
        raise RuntimeError("Cannot find {}".format(resfile))

    print("\nid22sumpy: processing file " + filename)

    if filename.endswith(ascii_extension):
        specfile = filename
        scan_numbers = specutils.saved_scan_numbers(specfile)
        if entries:
            scan_numbers = sorted(set(scan_numbers) - set(map(int, entries)))
    else:
        # Convert HDF5 to ASCII
        scans_h5 = get_all_scans_h5(filename)
        if not scans_h5:
            print("id22sumpy: No scans to process")
            return

        specfile = os.path.splitext(filename)[0] + ascii_extension
        specfile = os.path.join(outdir, os.path.basename(specfile))

        processed = specutils.saved_scan_numbers(specfile)

        scan_numbers = list()
        for entry in scans_h5:
            if not entry.endswith(".1"):
                continue
            if entries and entry not in entries:
                continue
            scannr = int(float(entry))
            if scannr in processed:
                continue
            data = read_h5(filename, entry)
            monav = numpy.mean(data["norm"], axis=0)
            save_spec(specfile, data, monav, entry)
            scan_numbers.append(scannr)

    if not scan_numbers:
        print("id22sumpy: No scans to process")
        return specfile

    first_scan = min(scan_numbers)
    last_scan = max(scan_numbers)
    missing_scans = sorted(set(range(first_scan, last_scan + 1)) - set(scan_numbers))
    gen_es = generate_es(first_scan, last_scan, excludescan, missing_scans)

    if zingit:
        program = "zingit"
        if programdir:
            program = os.path.join(programdir)
        cmd = [program, specfile, resbasename]
        execute_cmd(cmd, outdir)

    sbinsize = str(binsize)

    execute_sum(
        programdir,
        specfile,
        sbinsize,
        first_scan,
        last_scan,
        lowtth,
        scaling,
        zingit,
        sum_single,
        sum_all,
        excludedetector,
        gen_es,
        outdir,
        advanced=advanced,
        interactive=interactive,
    )

    while interactive and not yes_no(
        "Are you happy with selected binsize and scans[y/n]? "
    ):
        if yes_no("Do you want to change binsize[y/n]? "):
            sbinsize = new_bin(
                "Please type new binsize for {} [Ex:0.002]: ".format(specfile)
            )
        if yes_no("Do you want to exclude specific scans[y/n]? "):
            es_new = removed_scans(
                "Please type scan(s) to be removed from {} [Ex:2 or 1,2,5]: ".format(
                    specfile
                )
            )
            gen_es = generate_es(
                first_scan, last_scan, excludescan, missing_scans, es_new
            )

        execute_sum(
            programdir,
            specfile,
            sbinsize,
            first_scan,
            last_scan,
            lowtth,
            scaling,
            zingit,
            sum_single,
            sum_all,
            excludedetector,
            gen_es,
            outdir,
            interactive=interactive,
        )

    # Rename the sumall file to include the binsize
    xyfile = os.path.splitext(specfile)[0] + ".xye"
    if os.path.exists(xyfile):
        xybfile = os.path.splitext(xyfile)[0] + "_b{}.xye".format(
            sbinsize.replace(".", "")
        )
        os.rename(xyfile, xybfile)

    return specfile


def main(argv=None):
    import argparse
    from datetime import datetime

    if argv is None:
        argv = sys.argv

    starttime = datetime.now()

    parser = argparse.ArgumentParser(
        description="Script for processing of high resolution diffraction data using id22sume @ ID22, ESRF ",
        usage=usage_msg(),
        epilog=epilog_msg(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--filename",
        type=str,
        required=True,
        help="Full filename of a .h5-file, or start of filename for processing all files with same starting filename. Ex: hc4000_sample1_",
    )
    parser.add_argument(
        "-b", "--binsize", type=str, required=True, help="Binsize. Ex: 0.002"
    )
    parser.add_argument(
        "-lowtth", "--lowtth", type=float, default=0, help="Set low tth. Default is 0."
    )
    parser.add_argument(
        "-scal",
        "--scaling",
        type=str,
        default=None,
        help="Pattern scaling options (default is option 1 with default value for XXX): 1) scalmon=XXX: for counts per XXX monitor counts (default = 100000) 2) scalpk: to scale highest peak to correct number of obs counts 3) scaltot: to make total number of counts in scan correct 4) no: for no pattern scaling",
    )
    parser.add_argument(
        "-es",
        "--excludescan",
        type=str,
        help="Scans to be excluded. Ex: 3 or 3,5,8. Default is none.",
    )
    parser.add_argument(
        "-ed",
        "--excludedetector",
        type=str,
        help="Detectors to be excluded. Ex: 3 or 3,5,8. Default is none.",
    )
    parser.add_argument(
        "-int",
        "--interactive",
        type=int,
        default=0,
        help="Set to 1 if you want plotting and the option to change binsize and/or selected scans for each .dat-file being processed. Default is not interactive mode.",
    )
    parser.add_argument(
        "-all",
        "--sum_all",
        type=int,
        default=1,
        help="Set to 0 if you do NOT want to do id22sumall. Default is to do id22sumall.",
    )
    parser.add_argument(
        "-single",
        "--sum_single",
        type=int,
        default=1,
        help="Set to 0 if you do NOT want to do id22sum. Default is to do id22sum.",
    )
    parser.add_argument(
        "-zin",
        "--zingit",
        type=int,
        default=0,
        help="Set 1 if you need to do zingit (looking for saturated pixels in data). Default is not to do it. !!SHOULD NOT BE NEEDED ANYMORE!!",
    )
    parser.add_argument(
        "-a",
        "--advanced",
        type=str,
        default=None,
        nargs="+",
        help="For using any of the id22sum options not directly supported with this python-script. Ex: hightth=120 zap=1 gsas. Type id22sum+Enter for full list of options.",
    )
    parser.add_argument(
        "--programdir",
        type=str,
        default=None,
        help="Directory of the fortran programs.",
    )

    args = parser.parse_args(argv[1:])

    programdir = args.programdir
    if not programdir and os.path.exists("/users/opid22/bin"):
        programdir = "/users/opid22/bin"

    all_h5_files = find_h5_files(args.filename)
    if not all_h5_files:
        print(
            "Did not find any files starting/matching with {}...".format(args.filename)
        )

    for filename in all_h5_files:
        run(
            filename,
            binsize=args.binsize,
            lowtth=args.lowtth,
            scaling=args.scaling,
            excludescan=args.excludescan,
            excludedetector=args.excludedetector,
            interactive=args.interactive,
            sum_single=args.sum_single,
            sum_all=args.sum_all,
            zingit=args.zingit,
            advanced=args.advanced,
            programdir=args.programdir,
        )

    print(
        "Finished! :) Time taken in hh:min:sec : {} ".format(datetime.now() - starttime)
    )


if __name__ == "__main__":
    sys.exit(main())



import os
import shutil
from ewokscore import Task
from .id22sum import run
from . import dirutils


class ID22Sum(
    Task,
    input_names=["filename", "binsize"],
    optional_input_names=[
        "entries",
        "lowtth",
        "scaling",
        "sum_single",
        "sum_all",
        "advanced",
        "resfile",
        "outdirs",
        "primary_outdir",
        "outprefix",
        "raw_filename",
        "ascii_extension",
    ],
    output_names=["outfile"],
):
    def run(self):
        # TODO: only outdirs["primary"] is used

        options = {
            "filename": self.inputs.filename,
            "entries": self.inputs.entries,
            "binsize": self.inputs.binsize,
            "lowtth": self.get_input_value("lowtth", 0),
            "scaling": self.get_input_value("scaling", None),
            "sum_single": self.get_input_value("sum_single", 1),
            "sum_all": self.get_input_value("sum_all", 1),
            "advanced": self.get_input_value("advanced", None),
            "ascii_extension": self.get_input_value("ascii_extension", ".dat"),
        }
        do_sum = options["sum_single"] or options["sum_all"]
        if not do_sum:
            self.outputs.outfile = None
            return

        if self.missing_inputs.raw_filename:
            primary_outdir = None
        else:
            primary_outdir = self.get_input_value("primary_outdir", None)

        outdirs = self.get_input_value("outdirs", dict())
        outdirs = dirutils.prepare_outdirs(outdirs, primary_outdir)

        # Save results in a subdirectory
        if do_sum:
            prefix = os.path.splitext(os.path.basename(self.inputs.filename))[0]
            outdirs_fortran = dict()
            for name, (userhost, dirname) in outdirs.items():
                sbinsize = str(options["binsize"]).replace(".", "")
                subdirname = f"{prefix}_b{sbinsize}"
                if not self.missing_inputs.outprefix:
                    subdirname = self.inputs.outprefix + "_" + subdirname
                dirname = os.path.join(dirname, subdirname)
                outdirs_fortran[name] = userhost, dirname
            options["outdir"] = outdirs_fortran["primary"][1]
        else:
            options["outdir"] = outdirs["primary"][1]

        if do_sum and not self.missing_inputs.resfile:
            # The fortran rebin program will be executed in outdir.
            # It needs the file res-file in that directory.
            resbasename = os.path.basename(self.inputs.resfile)
            options["resbasename"] = resbasename
            resfile = dirutils.primary_file(resbasename, outdirs_fortran)
            dirname = os.path.dirname(resfile)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            shutil.copyfile(self.inputs.resfile, resfile)

        print(f"Result will be saved in: {options['outdir']}")

        self.outputs.outfile = run(**options)



import os
import shutil
import socket
import subprocess
from typing import Dict, Optional, Tuple, Iterator

HOST_NAME = socket.gethostname()

PARSED_OUTDIR = Tuple[Optional[str], str]
PARSED_OUTDIRS = Dict[str, PARSED_OUTDIR]


def prepare_outdirs(outdirs: Dict[str, str], primary_outdir: str) -> PARSED_OUTDIRS:
    """
    :param outdirs: maps names to local directories (/users/opid22/data1/)
                    or remote directories (opid22@diffract22new:/users/opid22/data1/)
    :param primary_outdir:
    """
    if outdirs is None:
        outdirs = dict()
    else:
        outdirs = dict(outdirs)
    if primary_outdir:
        for key in _outdir_key_generator():
            if key not in outdirs:
                break
        outdirs[key] = primary_outdir
    outdirs = {name: parse_outdir(dirname) for name, dirname in outdirs.items()}
    if "primary" not in outdirs:
        return dict()
    userhost, _ = outdirs["primary"]
    if userhost:
        raise ValueError("The primary output directory should be a local directory")
    return outdirs


def _outdir_key_generator() -> Iterator[str]:
    yield "primary"
    yield "processed"
    i = 0
    while True:
        i += 1
        yield f"processed{i}"


def parse_outdir(dirname: str) -> PARSED_OUTDIR:
    err_msg = f"malformed directory name '{dirname}'"
    if dirname.count(":") > 1:
        raise ValueError(err_msg)
    parts = dirname.split(":")
    if len(parts) not in (1, 2):
        raise ValueError(err_msg)
    if len(parts) == 1:
        return None, dirname
    userhost, dirname = parts
    if userhost.endswith(HOST_NAME):
        return None, dirname
    return userhost, dirname


def copy_file(filename: str, outdirs: PARSED_OUTDIRS) -> None:
    """Copy file from the primary output directory to the others."""
    if not outdirs:
        return
    local_filename = primary_file(filename, outdirs)
    filename = os.path.basename(local_filename)
    for name, (userhost, dirname) in outdirs.items():
        if name == "primary":
            continue
        remote_filename = os.path.join(dirname, filename)
        if userhost:
            cmd = ["scp", "-q", local_filename, f"{userhost}:'{remote_filename}'"]
            output = subprocess.check_output(cmd)
            if output:
                print(output)
        else:
            dirname = os.path.dirname(remote_filename)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            shutil.copyfile(local_filename, remote_filename)


def copy_file_to_primary(filename: str, outdirs: PARSED_OUTDIRS) -> None:
    """Copy file to the primary output directory."""
    shutil.copyfile(filename, primary_file(filename, outdirs))


def primary_file(filename: str, outdirs: PARSED_OUTDIRS) -> str:
    if not outdirs:
        return
    filename = os.path.basename(filename)
    return os.path.join(outdirs["primary"][1], filename)
